/*****************************************************************************
 * Copyright  2018 NXP
 * All rights reserved
 *

 * SPDX-License-Identifier: LA_OPT_BASE_LICENSE v24 May 2018
 *****************************************************************************/
/*****************************************************************************
 *
 * MODULE:          JN-AN-1243 Base Device Application
 *
 * COMPONENT:       app_ntag_reset.c
 *
 * DESCRIPTION:     Application layer for NTAG (reset)
 *
 * This file is INCOMPLETE and EXPERIMENTAL.
 *
 * It is a work in progress to use the NTAG blocking functions for commissioning
 * and ota. So far it only reads and validates the commissioning data.
 *
 ****************************************************************************
 *
 * This software is owned by NXP B.V. and/or its supplier and is protected
 * under applicable copyright laws. All rights are reserved. We grant You,
 * and any third parties, a license to use this software solely and
 * exclusively on NXP products [NXP Microcontrollers such as JN5168, JN5169,
 * JN5179, JN5189].
 * You, and any third parties must reproduce the copyright and warranty notice
 * and any other legend of ownership on each copy or partial copy of the
 * software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright NXP B.V. 2017-2018. All rights reserved
 *
 ***************************************************************************/

/****************************************************************************/
/***        Include files                                                 ***/
/****************************************************************************/
#include "fsl_i2c.h"
#include "fsl_ntag.h"
#include "fsl_clock.h"
#include "ntag.h"
#include "app_ntag_reset.h"
#include "dbg.h"   

/****************************************************************************/
/***        Macro Definitions                                             ***/
/****************************************************************************/
#ifndef DEBUG_NTAG_RESET
#define DEBUG_NTAG_RESET FALSE
#endif
/****************************************************************************/
/***        Type Definitions                                              ***/
/****************************************************************************/
uint8_t ntagbuf[NTAG_BLOCK_SIZE];
i2c_master_config_t s_masterConfig;
#define NTAG_POWER_UP_DELAY_US  300
#define NTAG_I2C_BAUDRATE 		100000U
#define NTAG_I2C_RESET    		kFC6_RST_SHIFT_RSTn
/****************************************************************************
 *
 * NAME:       vNtagEnable
 *
 * DESCRIPTION:
 * Enabled (power up and initialise I2c) or Disable (power down) the internal NTAG
 *
 * RETURNS:
 * void
 *
 ****************************************************************************/
static void vNtagEnable(bool_t bOn)
{
    if (bOn)
    {
        /* Switch on NTAG, configure IO pads */
        NTAG_SetState(kNTAG_StateActive);

        /* Wait for NTAG to be ready */
        CLOCK_uDelay(NTAG_POWER_UP_DELAY_US);

        /* Reset NTAG I2C */
        RESET_PeripheralReset(NTAG_I2C_RESET);

        /*
         * masterConfig.debugEnable = false;
         * masterConfig.ignoreAck = false;
         * masterConfig.pinConfig = kI2C_2PinOpenDrain;
         * masterConfig.baudRate_Bps = 100000U;
         * masterConfig.busIdleTimeout_ns = 0;
         * masterConfig.pinLowTimeout_ns = 0;
         * masterConfig.sdaGlitchFilterWidth_ns = 0;
         * masterConfig.sclGlitchFilterWidth_ns = 0;
         */
        I2C_MasterGetDefaultConfig(&s_masterConfig);

        /* Change the default baudrate configuration */
        s_masterConfig.baudRate_Bps = NTAG_I2C_BAUDRATE;

        /* Initialize the I2C master peripheral */
        I2C_MasterInit(((I2C_Type *) I2C2_BASE), &s_masterConfig, CLOCK_GetFreq(kCLOCK_Fro32M));
    }
    else
    {
        /* Disable I2C master */
        I2C_MasterDeinit(((I2C_Type *) I2C2_BASE));

        /* Switch off NTAG, configure IO pads */
        NTAG_SetState(kNTAG_StateInactive);
    }
}

/****************************************************************************
 *
 * NAME:       iNTAG_I2C_Write
 *
 * DESCRIPTION:
 * Perform I2c Write transfer to NTAG
 *
 * RETURNS:
 * int - returns 0 if OK
 *
 ****************************************************************************/
static int iNTAG_I2C_Write(uint32_t slvAddr, uint8_t *data, size_t len)
{
    i2c_master_transfer_t masterXfer;
    status_t status;
    int ret = 1;

    /* Initialise Transfer structure to default */
    memset(&masterXfer, 0, sizeof(masterXfer));

    /* Initialise Transfer structure with user data */
    masterXfer.slaveAddress = slvAddr;
    masterXfer.direction = kI2C_Write;
    masterXfer.subaddress = 0;
    masterXfer.subaddressSize = 0;
    masterXfer.data = data;
    masterXfer.dataSize = len;
    masterXfer.flags = kI2C_TransferDefaultFlag;
    status = I2C_MasterTransferBlocking(((I2C_Type *) I2C2_BASE), &masterXfer);

    if (status == kStatus_Success)
    {
        ret = 0;
    }
    else {
            DBG_vPrintf(DEBUG_NTAG_RESET,"I2C Error :%d\r\n",status);
    }

    return ret;
}

/****************************************************************************
 *
 * NAME:       iNTAG_I2C_Read
 *
 * DESCRIPTION:
 * Perform I2c Read transfer to NTAG
 *
 * RETURNS:
 * int - returns 0 if OK
 *
 ****************************************************************************/
static int iNTAG_I2C_Read(uint32_t slvAddr, uint8_t *data, size_t len)
{
    i2c_master_transfer_t masterXfer;
    status_t status;
    int ret = 1;

    /* Initialise Transfer structure to default */
    memset(&masterXfer, 0, sizeof(masterXfer));

    /* Initialise Transfer structure with user data */
    masterXfer.slaveAddress = slvAddr;
    masterXfer.direction = kI2C_Read;
    masterXfer.subaddress = 0;
    masterXfer.subaddressSize = 0;
    masterXfer.data = data;
    masterXfer.dataSize = len;
    masterXfer.flags = kI2C_TransferDefaultFlag;
    status = I2C_MasterTransferBlocking(((I2C_Type *) I2C2_BASE), &masterXfer);

    if (status == kStatus_Success)
    {
        ret = 0;
    }

    return ret;
}

/****************************************************************************
 *
 * NAME:       iNTAG_ReadBlock
 *
 * DESCRIPTION:
 * Function to read a block of NTAG EEPROM over I2C
 *
 * RETURNS:
 * int - returns 0 if OK
 *
 ****************************************************************************/
PUBLIC int iNTAG_ReadBlock(uint32_t slvAddr, uint32_t blockid, uint8_t *data)
{
    uint8_t frame[1];
    int ret;

    frame[0] = (uint8_t) blockid;

    ret = iNTAG_I2C_Write(slvAddr, frame, sizeof(frame));

    if (ret == 0)
    {
        ret = iNTAG_I2C_Read(slvAddr, data, NTAG_BLOCK_SIZE);
    }

    return ret;
}
/****************************************************************************
 *
 * NAME:       iNTAG_WriteBlock
 *
 * DESCRIPTION:
 * Function to write a block of NTAG EEPROM over I2C
 *
 * RETURNS:
 * int - returns 0 if OK
 *
 ****************************************************************************/
PUBLIC int iNTAG_WriteBlock(uint32_t slvAddr, uint32_t blockid, const uint8_t *data)
{
    uint8_t frame[NTAG_BLOCK_SIZE + 1];
    int ret;

    frame[0] = (uint8_t) blockid;
    memcpy(&frame[1], data, NTAG_BLOCK_SIZE);

    ret = iNTAG_I2C_Write(slvAddr, frame, sizeof(frame));
    return ret;
}
/****************************************************************************
 *
 * NAME:       iNTAG_DumpBlock
 *
 * DESCRIPTION:
 * Function to dump a block of NTAG EEPROM to the UART
 *
 * RETURNS:
 * int - returns 0 if OK
 *
 ****************************************************************************/
PUBLIC int iNTAG_DumpBlock(uint32_t u32BlockId, uint32_t slvAddr)
{
    int ret = -1;
    size_t i;

    if (iNTAG_ReadBlock(slvAddr, u32BlockId, ntagbuf) == 0)
    {
    	if(u32BlockId==0){
    			DBG_vPrintf(DEBUG_NTAG_RESET,"Found NTAG I2C address: 0x%x\n\r", slvAddr);
    	}
        DBG_vPrintf(DEBUG_NTAG_RESET,"Block %3d: ",u32BlockId);
        for (i = 0; i < NTAG_BLOCK_SIZE; i++)
        {
            DBG_vPrintf(DEBUG_NTAG_RESET,"%02x ", ntagbuf[i]);
        }
        DBG_vPrintf(DEBUG_NTAG_RESET,"\n\r");

        ret = 0;
    }

    return ret;
}
/****************************************************************************
 *
 * NAME:       vResetNTAGAddress
 *
 * DESCRIPTION:
 * Re-programs NTAG slave address to 0x55 if accidentally reprogrammed
 *
 * Value always read as 0x4 so Reading and writing the same values back will
 * change the address
 *
 * RETURNS:
 * void
 *
 ****************************************************************************/
PUBLIC void vResetNTAGAddress(void ){
	int address;
	DBG_vPrintf(DEBUG_NTAG_RESET,"\n\r");
       vNtagEnable(TRUE);
	for (address = 0; address <= 127; address++)
	{
		if (iNTAG_DumpBlock(0,address) == 0)
		{
			if (address!=0x55){
				ntagbuf[0] = 0xAA;
				iNTAG_WriteBlock(address, 0 , ntagbuf);
				DBG_vPrintf(DEBUG_NTAG_RESET,"I2C address re-programmed\n\r");
			}
			else{
				DBG_vPrintf(DEBUG_NTAG_RESET,"I2C address already set to 0x55\n\r");
			}
			break;
		}
	}

	if (address > 127) {
		DBG_vPrintf(DEBUG_NTAG_RESET,"Error - no NTAG detected\n\r");
	}
}
/****************************************************************************/
/***        Exported Variables                                            ***/
/****************************************************************************/

/****************************************************************************/
/***        END OF FILE                                                   ***/
/****************************************************************************/
