/*****************************************************************************
 * Copyright  2018 NXP
 * All rights reserved
 *

 * SPDX-License-Identifier: LA_OPT_BASE_LICENSE v24 May 2018
 *****************************************************************************/
/*****************************************************************************
 *
 * MODULE:             Motor Driver
 *
 * COMPONENT:          MotorDriver.c
 *
 * AUTHOR:             Martin Looker
 *
 * DESCRIPTION:        Header for A4988 Stepper Motor Driver
 *
 *****************************************************************************
 *
 * This software is owned by NXP B.V. and/or its supplier and is protected
 * under applicable copyright laws. All rights are reserved. We grant You,
 * and any third parties, a license to use this software solely and
 * exclusively on NXP products  [NXP Microcontrollers such as JN5189, JN%179,
 * JN5168, JN5164, JN5161].
 * You, and any third parties must reproduce the copyright and warranty notice
 * and any other legend of ownership on each  copy or partial copy of the software.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright NXP B.V. 2019. All rights reserved
 *
 ****************************************************************************/
#include <jendefs.h>
#include <string.h>

#ifndef K32W148EVK
#include "fsl_iocon.h"
#include "fsl_gpio.h"
#endif

#include "dbg.h"

#include "DriverMotor.h"

/****************************************************************************/
/***        Macro Definitions                                             ***/
/****************************************************************************/

/****************************************************************************/
/***        Type Definitions                                              ***/
/****************************************************************************/

/****************************************************************************/
/***        Local Function Prototypes                                     ***/
/****************************************************************************/
PRIVATE bool_t bDM_Forward(void);
PRIVATE bool_t bDM_Backward(void);
#ifdef DM_ULN2003
PRIVATE void vDM_Phase(void);
#endif

/****************************************************************************/
/***        Exported Variables                                            ***/
/****************************************************************************/

/****************************************************************************/
/***        Local Variables                                               ***/
/****************************************************************************/
PRIVATE tsDM_State *psDmState; /* Pointer to state data */
PRIVATE uint32     u32DmStepTarget = 0xFFFFFFFF;

/****************************************************************************/
/***        Exported Functions                                            ***/
/****************************************************************************/
/****************************************************************************
 * NAME: vDM_Initialise
 *
 * DESCRIPTION:
 * Initialises motor driver
 *
 * PARAMETERS:  Name                            Usage
 *              psDmStateInit                   Pointer to state data to use
 *
 * RETURNS:
 *  void
 ****************************************************************************/
PUBLIC void vDM_Initialise(tsDM_State *psDmStateInit)
{
    gpio_pin_config_t output_config = {kGPIO_DigitalOutput, OFF};

    /* Note state pointer */
    psDmState = psDmStateInit;
    /* Set direction to stopped */
    psDmState->eDir = E_DM_DIR_STOP;
    /* Validate soft limits against hard limits */
    if (psDmState->u32StepSoftMin < DM_STEP_HARD_MIN) psDmState->u32StepSoftMin = DM_STEP_HARD_MIN;
    if (psDmState->u32StepSoftMax > DM_STEP_HARD_MAX) psDmState->u32StepSoftMax = DM_STEP_HARD_MAX;
    /* ULN2003 ? */
    #ifdef DM_ULN2003
    {
 #ifndef K32W148EVK
        uint32 u32Mask;

        /* Set muxing for output lines */
        IOCON_PinMuxSet(IOCON, 0, DM_OUTPUT_PIN_A,  IOCON_FUNC0 | IOCON_DIGITAL_EN | IOCON_GPIO_MODE | IOCON_MODE_INACT);
        IOCON_PinMuxSet(IOCON, 0, DM_OUTPUT_PIN_B,  IOCON_FUNC0 | IOCON_DIGITAL_EN | IOCON_GPIO_MODE | IOCON_MODE_INACT);
        IOCON_PinMuxSet(IOCON, 0, DM_OUTPUT_PIN_C,  IOCON_FUNC0 | IOCON_DIGITAL_EN | IOCON_GPIO_MODE | IOCON_MODE_INACT);
        IOCON_PinMuxSet(IOCON, 0, DM_OUTPUT_PIN_D,  IOCON_FUNC0 | IOCON_DIGITAL_EN | IOCON_GPIO_MODE | IOCON_MODE_INACT);
        /* Configure as output pins (defaulting to off) */
        GPIO_PinInit(GPIO, APP_BOARD_GPIO_PORT, DM_OUTPUT_PIN_A, &output_config);
        GPIO_PinInit(GPIO, APP_BOARD_GPIO_PORT, DM_OUTPUT_PIN_B, &output_config);
        GPIO_PinInit(GPIO, APP_BOARD_GPIO_PORT, DM_OUTPUT_PIN_C, &output_config);
        GPIO_PinInit(GPIO, APP_BOARD_GPIO_PORT, DM_OUTPUT_PIN_D, &output_config);
        /* Set mask to all outputs */
        u32Mask = ~((1<<DM_OUTPUT_PIN_A)|(1<<DM_OUTPUT_PIN_B)|(1<<DM_OUTPUT_PIN_C)|(1<<DM_OUTPUT_PIN_D));
        /* Add outputs to mask */
        GPIO_PortMaskedSet(GPIO, APP_BOARD_GPIO_PORT, u32Mask);
        /* Go to phase for step */
        vDM_Phase();
#endif
    }
    #endif
}

/****************************************************************************
 * NAME: eDM_Stop
 *
 * DESCRIPTION:
 * Stops moving motor
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  teDM_Dir - current direction (always STOP)
 ****************************************************************************/
PUBLIC teDM_Dir eDM_Stop(void)
{
    /* Set mode to stopped */
    psDmState->eDir = E_DM_DIR_STOP;
    /* Invalidate step target */
    u32DmStepTarget = 0xFFFFFFFF;

    /* Return */
    return psDmState->eDir;
}

/****************************************************************************
 * NAME: eDM_Forward
 *
 * DESCRIPTION:
 * Starts moving motor forward
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  teDM_Dir - current direction (FORWARD if successful)
 ****************************************************************************/
PUBLIC teDM_Dir eDM_Forward(void)
{
    /* OK to move forward ? */
    if (bDM_Forward())
    {
        /* Not already going forwards ? */
        if (psDmState->eDir != E_DM_DIR_FORWARD)
        {
            /* Set mode to forward */
            psDmState->eDir = E_DM_DIR_FORWARD;
            /* Increment activations */
            psDmState->u32Activations++;
        }
    }

    /* Return */
    return psDmState->eDir;
}

/****************************************************************************
 * NAME: eDM_Backward
 *
 * DESCRIPTION:
 * Starts moving motor backwards
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  teDM_Dir - current direction (BACKWARD if successful)
 ****************************************************************************/
PUBLIC teDM_Dir eDM_Backward(void)
{
    /* OK to move backwards ? */
    if (bDM_Backward())
    {
        /* Not already going backwards ? */
        if (psDmState->eDir != E_DM_DIR_BACKWARD)
        {
            /* Set mode to backward */
            psDmState->eDir = E_DM_DIR_BACKWARD;
            /* Increment activations */
            psDmState->u32Activations++;
        }
    }

    /* Return */
    return psDmState->eDir;
}

/****************************************************************************
 * NAME: eDM_GoToTarget
 *
 * DESCRIPTION:
 * Moves motor to target step (only if valid)
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  teDM_Dir - current direction (BACKWARD if successful)
 ****************************************************************************/
PUBLIC teDM_Dir eDM_GoToTarget(uint32 u32StepTarget, bool_t *pbValid)
{
    /* Assume invalid */
    *pbValid = FALSE;
    /* Not in calibration mode ? */
    if ((psDmState->u8Mode & DM_MODE_BIT_CALIBRATION) == 0)
    {
        /* Is target in valid range ? */
        if (u32StepTarget >= psDmState->u32StepSoftMin && u32StepTarget <= psDmState->u32StepSoftMax)
        {
            /* Valid */
            *pbValid = TRUE;
        }
    }
    /* In calibration mode ? */
    else
    {
        /* Is target in valid range ? */
        if (u32StepTarget >= DM_STEP_HARD_MIN && u32StepTarget <= DM_STEP_HARD_MAX)
        {
            /* Valid */
            *pbValid = TRUE;
        }
    }
    /* Target is valid ? */
    if (*pbValid)
    {
        /* Need to move forward ? */
        if (psDmState->u32Step < u32StepTarget)
        {
            /* Set target */
            u32DmStepTarget = u32StepTarget;
            /* Attempt to move forward */
            eDM_Forward();
            /* Not moving forward ? */
            if (psDmState->eDir != E_DM_DIR_FORWARD)
            {
                /* Invalidate target */
                u32DmStepTarget = 0xFFFFFFFF;
            }
        }
        /* Need to move backward ? */
        else if (psDmState->u32Step > u32StepTarget)
        {
            /* Set target */
            u32DmStepTarget = u32StepTarget;
            /* Attempt to move backward */
            eDM_Backward();
            /* Not moving backward ? */
            if (psDmState->eDir != E_DM_DIR_BACKWARD)
            {
                /* Invalidate target */
                u32DmStepTarget = 0xFFFFFFFF;
            }
        }
    }

    /* Return */
    return psDmState->eDir;
}

/****************************************************************************
 * NAME: teDM_Tick
 *
 * DESCRIPTION:
 * Timer tick function for motor operation
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  teDM_Dir - current direction
 ****************************************************************************/
PUBLIC teDM_Dir eDM_Tick(void)
{
    /* Going forwards ? */
    if (psDmState->eDir == E_DM_DIR_FORWARD)
    {
        /* Ok to move forward ? */
        if (bDM_Forward())
        {
            /* Increment current step */
            psDmState->u32Step++;
        }
        /* Not ok to move forward ? */
        else
        {
            /* Stop */
            psDmState->eDir = E_DM_DIR_STOP;
            /* Invalidate step target */
            u32DmStepTarget = 0xFFFFFFFF;
        }
    }
    /* Going backwards ? */
    else if (psDmState->eDir == E_DM_DIR_BACKWARD)
    {
        /* Ok to move backward ? */
        if (bDM_Backward())
        {
            /* Decrement current step */
            psDmState->u32Step--;
        }
        /* Not ok to move forward ? */
        else
        {
            /* Stop */
            psDmState->eDir = E_DM_DIR_STOP;
            /* Invalidate step target */
            u32DmStepTarget = 0xFFFFFFFF;
        }
    }

    /* ULN2003 ? */
    #ifdef DM_ULN2003
    {
        /* Set phase from step (or disable for maintenence mode) */
        vDM_Phase();
    }
    #endif

    /* Return */
    return psDmState->eDir;
}

/****************************************************************************/
/***        Local Functions                                               ***/
/****************************************************************************/
/****************************************************************************
 * NAME: bDM_Forward
 *
 * DESCRIPTION:
 * Allowed to move forward ?
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  TRUE  - allowed
 *  FALSE - no allowed
 ****************************************************************************/
PRIVATE bool_t bDM_Forward(void)
{
    bool_t bReturn = FALSE;

    /* Not in maintenence mode ? */
    if ((psDmState->u8Mode & DM_MODE_BIT_MAINTENANCE) == 0)
    {
        /* Not in calibration mode ? */
        if ((psDmState->u8Mode & DM_MODE_BIT_CALIBRATION) == 0)
        {
            /* Below soft limit and not at target ? */
            if (psDmState->u32Step < psDmState->u32StepSoftMax && psDmState->u32Step != u32DmStepTarget)
            {
                /* Ok to move */
                bReturn = TRUE;
            }
        }
        /* Calibration mode ? */
        else
        {
            /* Below hard limit ? */
            if (psDmState->u32Step < DM_STEP_HARD_MAX && psDmState->u32Step != u32DmStepTarget)
            {
                /* Ok to move */
                bReturn = TRUE;
            }
        }
    }

    return bReturn;
}

/****************************************************************************
 * NAME: bDM_Backward
 *
 * DESCRIPTION:
 * Allowed to move backward ?
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  TRUE  - allowed
 *  FALSE - no allowed
 ****************************************************************************/
PRIVATE bool_t bDM_Backward(void)
{
    bool_t bReturn = FALSE;

    /* Not in maintenence mode ? */
    if ((psDmState->u8Mode & DM_MODE_BIT_MAINTENANCE) == 0)
    {
        /* Not in calibration mode ? */
        if ((psDmState->u8Mode & DM_MODE_BIT_CALIBRATION) == 0)
        {
            /* Above soft limit ? */
            if (psDmState->u32Step > psDmState->u32StepSoftMin && psDmState->u32Step != u32DmStepTarget)
            {
                /* Ok to move */
                bReturn = TRUE;
            }
        }
        /* Calibration mode ? */
        else
        {
            /* Above hard limit ? */
            if (psDmState->u32Step > DM_STEP_HARD_MIN && psDmState->u32Step != u32DmStepTarget)
            {
                /* Ok to move */
                bReturn = TRUE;
            }
        }
    }

    return bReturn;
}

#ifdef DM_ULN2003
/****************************************************************************
 * NAME: teDM_Phase
 *
 * DESCRIPTION:
 * Go to phase from step
 *
 * PARAMETERS:
 *  None
 *
 * RETURNS:
 *  teDM_Dir - current direction
 ****************************************************************************/
PRIVATE void vDM_Phase(void)
{
    uint8   u8Phase;
    uint32 u32Mask = 0;

    /* Not in maintenence mode ? */
    if ((psDmState->u8Mode & DM_MODE_BIT_MAINTENANCE) == 0)
    {
        /* Calculate phase from step */
        u8Phase = psDmState->u32Step & 0x3;
        /* Motor direction is normal */
        if ((psDmState->u8Mode & DM_MODE_BIT_REVERSED) == 0)
        {
            /* Set mask for phase */
            switch (u8Phase)
            {
                case 3:  u32Mask = (1 << DM_OUTPUT_PIN_D); break;
                case 2:  u32Mask = (1 << DM_OUTPUT_PIN_C); break;
                case 1:  u32Mask = (1 << DM_OUTPUT_PIN_B); break;
                default: u32Mask = (1 << DM_OUTPUT_PIN_A); break;
            }
        }
        /* Motor direction is reversed ? */
        else
        {
            /* Set mask for phase */
            switch (u8Phase)
            {
                case 3:  u32Mask = (1 << DM_OUTPUT_PIN_B); break;
                case 2:  u32Mask = (1 << DM_OUTPUT_PIN_C); break;
                case 1:  u32Mask = (1 << DM_OUTPUT_PIN_D); break;
                default: u32Mask = (1 << DM_OUTPUT_PIN_A); break;
            }
        }
    }
    /* Set masked outputs */
#ifndef K32W148EVK
    GPIO_PortMaskedWrite(GPIO, APP_BOARD_GPIO_PORT, u32Mask);
#endif
}
#endif

/****************************************************************************/
/***        END OF FILE                                                   ***/
/****************************************************************************/

