/*****************************************************************************
 * Copyright  2018 NXP
 * All rights reserved
 *

 * SPDX-License-Identifier: LA_OPT_BASE_LICENSE v24 May 2018
 *****************************************************************************/
/*****************************************************************************
 *
 * MODULE:      NFC
 *
 * COMPONENT:   nci_ota.h
 *
 * AUTHOR:      Martin Looker
 *
 * DESCRIPTION: NCI OTA reading and writing (interface)
 *
 * $HeadURL: https://www.collabnet.nxp.com/svn/lprf_sware/Projects/Components/NFC/Tags/+v1025/Include/nci_ota.h $
 *
 * $Revision: 102076 $
 *
 * $LastChangedBy: cnh01837 $
 *
 * $LastChangedDate: 2023-11-17 16:06:10 +0800 (周五, 17 11月 2023) $
 *
 * $Id: nci_ota.h 102076 2023-11-17 08:06:10Z cnh01837 $
 *
 ****************************************************************************
 *
 * This software is owned by NXP B.V. and/or its supplier and is protected
 * under applicable copyright laws. All rights are reserved. We grant You,
 * and any third parties, a license to use this software solely and
 * exclusively on NXP products [NXP Microcontrollers such as JN5168, JN5179].
 * You, and any third parties must reproduce the copyright and warranty notice
 * and any other legend of ownership on each copy or partial copy of the
 * software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright NXP B.V. 2017-2018. All rights reserved
 *
 ***************************************************************************/
/*!
 * \file   nci_ota.h
 *
 * \brief NCI OTA reading and writing (interface)
 */
/* nci_ota.h contains high level APIs for reading and writing OTA data via an NTAG's SRAM when
 * placed into the reader's field.
 *
 ***************************************************************************/
#ifndef NCI_OTA_H_
#define NCI_OTA_H_

/****************************************************************************/
/***        Include Files                                                 ***/
/****************************************************************************/
#include <jendefs.h>
#include "nfc_ota.h"

/****************************************************************************/
/***        Macro Definitions                                             ***/
/****************************************************************************/

/****************************************************************************/
/***        Type Definitions                                              ***/
/****************************************************************************/
/*! NCI OTA status type */
typedef enum
{
    E_NCI_OTA_IDLE,        /*!< NCI OTA processing is idle */
    E_NCI_OTA_READING,     /*!< NCI OTA is reading */
    E_NCI_OTA_READ_FAIL,   /*!< NCI OTA read has failed */
    E_NCI_OTA_READ_OK,     /*!< NCI OTA read was successful */
    E_NCI_OTA_WRITING,     /*!< NCI OTA is writing */
    E_NCI_OTA_WRITE_FAIL,  /*!< NCI OTA write has failed */
    E_NCI_OTA_WRITE_OK     /*!< NCI OTA write was successful */
} teNciOtaStatus;

/****************************************************************************/
/***        Exported Functions                                            ***/
/****************************************************************************/
/****************************************************************************
 *
 *  NAME:  NCI_OTA_eRead
 */
/*! \brief Requests read of OTA data from an NTAG's SRAM in the
 * reader's field
 */
/* If the request is successful the final outcome of the read request is
 * returned by NCI_OTA_eTick() returning a status of E_NCI_OTA_READ_OK.
 *
 * When the request is accepted the NCI event callback function, set by
 * NCI_vRegCbEvent(), is overridden to allow the NCI OTA code to process
 * NCI events. When the request completes NCI_OTA_eStop() should be called
 * to end the processing and the NTAG event callback function can be reclaimed
 * by the application.
 *
 * \retval E_NCI_OTA_READING   Request accepted
 * \retval E_NCI_OTA_READ_FAIL Request failed
 *
 ****************************************************************************/
PUBLIC teNciOtaStatus NCI_OTA_eRead(
    uint8 *pu8Read /*!< Pointer to read data into (64 bytes will be read) */
    );

/****************************************************************************
 *
 *  NAME:  NCI_OTA_eWrite
 */
/*! \brief Requests write of OTA data via NTAG's SRAM in the
 * reader's field
 */
/* If the request is successful the final outcome of the write request is
 * returned by NCI_OTA_eTick() returning a status of E_NCI_OTA_WRITE_OK.
 *
 * When the request is accepted the NCI event callback function, set by
 * NCI_vRegCbEvent(), is overridden to allow the NCI OTA code to process
 * NCI events. When the request completes NCI_OTA_eStop() should be called
 * to end the processing and the NCI event callback function can be reclaimed
 * by the application.
 *
 * \retval E_NCI_OTA_WRITING    Request accepted
 * \retval E_NCI_OTA_WRITE_FAIL Request failed
 *
 ****************************************************************************/
PUBLIC teNciOtaStatus NCI_OTA_eWrite(
    uint8 *pu8Write  /*!< Pointer to data to write (64 bytes will be written) */
    );

/****************************************************************************
 *
 *  NAME:  NCI_OTA_eStop
 */
/*! \brief Stops processing of NCI OTA data
 */
/* This function should be called when the processing of the OTA
 * data is complete, the NCI event callback function can be reclaimed by the
 * application after calling NCI_OTA_eStop().
 *
 * \retval E_NCI_OTA_IDLE NCI_OTA processing is idle
 *
 ****************************************************************************/
PUBLIC teNciOtaStatus NCI_OTA_eStop(void);

/****************************************************************************
 *
 *  NAME:  NCI_OTA_eStatus
 */
/*! \brief Returns the status of the NCI OTA processing
 */
/* \returns Status of NCI OTA processing
 *
 ****************************************************************************/
PUBLIC teNciOtaStatus NCI_OTA_eStatus(void);

/****************************************************************************
 *
 *  NAME:  NCI_OTA_eTick
 */
/*! \brief Timer function to drive NCI OTA processing
 */
/* Should be called regularly, every 5ms is recommended.
 *
 * \warning This function calls NCI_bTick() internally, there is no need to
 * call NCI_bTick() from the application when NCI OTA processing is
 * taking place.
 *
 * \returns Status of NCI OTA processing
 *
 ****************************************************************************/
PUBLIC teNciOtaStatus NCI_OTA_eTick(
    uint32               u32TickMs /*!< Time in ms since previous call */
    );

/****************************************************************************/
/***        Exported Variables                                            ***/
/****************************************************************************/

#endif /* NCI_OTA_H_ */
/****************************************************************************/
/***        END OF FILE                                                   ***/
/****************************************************************************/
