/*****************************************************************************
 * Copyright  2018 NXP
 * All rights reserved
 *

 * SPDX-License-Identifier: LA_OPT_BASE_LICENSE v24 May 2018
 *****************************************************************************/
/*****************************************************************************
 *
 * MODULE:      NFC
 *
 * COMPONENT:   nfc_nwk.h
 *
 * AUTHOR:      Martin Looker
 *
 * DESCRIPTION: Common macros used by all NFC NWK NDEF processing
 *
 * $HeadURL: https://www.collabnet.nxp.com/svn/lprf_sware/Projects/Components/NFC/Tags/+v1025/Include/nfc_nwk.h $
 *
 * $Revision: 102076 $
 *
 * $LastChangedBy: cnh01837 $
 *
 * $LastChangedDate: 2023-11-17 16:06:10 +0800 (周五, 17 11月 2023) $
 *
 * $Id: nfc_nwk.h 102076 2023-11-17 08:06:10Z cnh01837 $
 *
 ****************************************************************************
 *
 * This software is owned by NXP B.V. and/or its supplier and is protected
 * under applicable copyright laws. All rights are reserved. We grant You,
 * and any third parties, a license to use this software solely and
 * exclusively on NXP products [NXP Microcontrollers such as JN5168, JN5179].
 * You, and any third parties must reproduce the copyright and warranty notice
 * and any other legend of ownership on each copy or partial copy of the
 * software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright NXP B.V. 2016-2018. All rights reserved
 *
 ***************************************************************************/
/*!
 * \file  nfc_nwk.h
 *
 * \brief Common macros and functions used by all NFC NWK NDEF processing
 *
 ***************************************************************************/
#ifndef NFC_NWK_H_
#define NFC_NWK_H_

/****************************************************************************/
/***        Include Files                                                 ***/
/****************************************************************************/
#include <jendefs.h>

/****************************************************************************/
/***        Macro Definitions                                             ***/
/****************************************************************************/
/* NDEF defines */
#define NFC_NWK_NDEF_TYPE_LEN               18  /*!< Size of our NDEF type string (not terminated) */
/* Payload defines */
#define NFC_NWK_PAYLOAD_VERSION             13  /*!< NFC NWK NDEF payload version number */
#define NFC_NWK_PAYLOAD_KEY_SIZE            16  /*!< Size of payload key byte arrays */
#define NFC_NWK_PAYLOAD_MIC_SIZE             4  /*!< Size of payload MIC byte arrays */

/****************************************************************************/
/***        Type Definitions                                              ***/
/****************************************************************************/
#pragma GCC diagnostic ignored "-Wpacked"
#pragma GCC diagnostic ignored "-Wattributes"

/*! Data written by NTAG side (44 bytes) */
typedef struct
{
    uint8           u8Version;                          /*!< ( 1) Data structure version */
    uint8           u8Command;                          /*!< ( 1) Command from NTAG to NCI */
    uint8           u8Sequence;                         /*!< ( 1) Write counter */
    uint16          u16DeviceId;                        /*!< ( 2) Device type ID */
    uint64          u64ExtAddress;                      /*!< ( 8) Extended address */
    uint16          u16ShortAddress;                    /*!< ( 2) Current short address */
    uint8           u8Channel;                          /*!< ( 1) Current channel */
    uint16          u16PanId;                           /*!< ( 2) Current PAN ID */
    uint64          u64ExtPanId;                        /*!< ( 8) Current extended PAN ID */
    uint8           au8Key[NFC_NWK_PAYLOAD_KEY_SIZE];   /*!< (16) Installation code */
    uint16          u16Crc;                             /*!< ( 2) Installation code CRC */
} PACK tsNfcNwkNtag;                                    /*   (44) */

/*! Data written by NCI side (46 bytes) */
typedef struct
{
    uint8           u8Command;                          /*!< ( 1) Command from NCI to NTAG */
    uint8           u8Sequence;                         /*!< ( 1) Write counter */
    uint16          u16DeviceId;                        /*!< ( 2) Device type ID */
    uint64          u64ExtAddress;                      /*!< ( 8) Network trust center extended address */
    uint16          u16ShortAddress;                    /*!< ( 2) Network trust center short address */
    uint8           u8Channel;                          /*!< ( 1) Network channel */
    uint16          u16PanId;                           /*!< ( 2) Network PAN ID */
    uint64          u64ExtPanId;                        /*!< ( 8) Network extended PAN ID */
    uint8           au8Key[NFC_NWK_PAYLOAD_KEY_SIZE];   /*!< (16) Encrypted network key */
    uint8           au8Mic[NFC_NWK_PAYLOAD_MIC_SIZE];   /*!< ( 4) MIC for encrypted key */
    uint8           u8KeySeqNum;                        /*!< ( 1) Network key sequence number */
} PACK tsNfcNwkNci;                                     /*   (46) */

/*! Payload data (90 bytes) */
typedef struct
{
    tsNfcNwkNtag    sNtag;                              /*!< (44) Data written by ntag side */
    tsNfcNwkNci     sNci;                               /*!< (46) Data written by nci side */
} PACK tsNfcNwkPayload;                                 /*   (90) */

/*! NDEF data (111 bytes) */
typedef struct
{
    uint8           u8NdefFlags;                        /*!< (  1) 0xd4 = MB=1, ME=1, CF=0, SR=1, IL=0, TNF=0x4 (External) */
    uint8           u8NdefTypeLength;                   /*!< (  1) NFC_NWK_NDEF_TYPE_LEN (18) */
    uint8           u8NdefPayloadLength;                /*!< (  1) sizeof(tsNfcNwkPayload) */
    uint8           au8NdefType[NFC_NWK_NDEF_TYPE_LEN]; /*!< ( 18) "com.nxp:JN51xx-NWK" */
    tsNfcNwkPayload sPayload;                           /*!< ( 90) NDEF payload */
} PACK tsNfcNwkNdef;                                    /*   (111) */

/*! TLV data (113 bytes) */
typedef struct
{
    uint8           u8TlvType;                          /*!< (  1) 0x03 = NDEF */
    uint8           u8TlvLength;                        /*!< (  1) sizeof(tsNfcNwkNdef) */
    tsNfcNwkNdef    sNdef;                              /*!< (111) NDEF structure */
} PACK tsNfcNwkTlv;                                     /*   (113) */

/*! EEPROM data (114 bytes) */
typedef struct
{
    tsNfcNwkTlv     sTlv;                               /*!< (113) NDEF structure */
    uint8           u8TlvTerminator;                    /*!< (  1) 0xfe = Terminator Tlv Type */
} PACK tsNfcNwk;                                        /*   (114) */

/*! Installation code structure */
typedef struct
{
    uint8  au8Key[NFC_NWK_PAYLOAD_KEY_SIZE];            /*!< Installation code */
    uint16 u16Crc;                                      /*!< Installation code CRC */
} PACK tsNfcNwkInstallCode;

#pragma GCC diagnostic pop

/****************************************************************************/
/***        Exported Functions                                            ***/
/****************************************************************************/
/****************************************************************************
 *
 *  NAME:  NFC_NWK_vNfcNwkFromBuffer
 */
/*! \brief Converts raw buffer data to tsNfcNwk data structure
 *
 ****************************************************************************/
PUBLIC void   NFC_NWK_vNfcNwkFromBuffer(
    tsNfcNwk  *psNfcNwk,  /*!< Pointer to structure for output data */
    uint8    *pu8Buffer   /*!< Pointer to buffer for input data */
    );

/****************************************************************************
 *
 *  NAME:  NFC_NWK_vNfcNwkToBuffer
 */
/*! \brief Converts tsNfcNwk data structure to raw data buffer
 *
 ****************************************************************************/
PUBLIC void   NFC_NWK_vNfcNwkToBuffer(
    tsNfcNwk  *psNfcNwk,  /*!< Pointer to structure for input data */
    uint8    *pu8Buffer   /*!< Pointer to buffer for output data */
    );

/****************************************************************************/
/***        Exported Variables                                            ***/
/****************************************************************************/

#endif /* NFC_NWK_H_ */
/****************************************************************************/
/***        END OF FILE                                                   ***/
/****************************************************************************/
