/*****************************************************************************
 * Copyright  2018 NXP
 * All rights reserved
 *

 * SPDX-License-Identifier: LA_OPT_BASE_LICENSE v24 May 2018
 *****************************************************************************/
/*****************************************************************************
 *
 * MODULE:      NFC
 *
 * COMPONENT:   nfc_ota.h
 *
 * AUTHOR:      Martin Looker
 *
 * DESCRIPTION: Common macros used by all NFC OTA processing
 *
 * $HeadURL: https://www.collabnet.nxp.com/svn/lprf_sware/Projects/Components/NFC/Tags/+v1025/Include/nfc_ota.h $
 *
 * $Revision: 102076 $
 *
 * $LastChangedBy: cnh01837 $
 *
 * $LastChangedDate: 2023-11-17 16:06:10 +0800 (周五, 17 11月 2023) $
 *
 * $Id: nfc_ota.h 102076 2023-11-17 08:06:10Z cnh01837 $
 *
 ****************************************************************************
 *
 * This software is owned by NXP B.V. and/or its supplier and is protected
 * under applicable copyright laws. All rights are reserved. We grant You,
 * and any third parties, a license to use this software solely and
 * exclusively on NXP products [NXP Microcontrollers such as JN5168, JN5179].
 * You, and any third parties must reproduce the copyright and warranty notice
 * and any other legend of ownership on each copy or partial copy of the
 * software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright NXP B.V. 2017-2018. All rights reserved
 *
 ***************************************************************************/
/*!
 * \file  nfc_ota.h
 *
 * \brief Common macros and functions used by all NFC OTA processing
 *
 ***************************************************************************/
#ifndef NFC_OTA_H_
#define NFC_OTA_H_

/****************************************************************************/
/***        Include Files                                                 ***/
/****************************************************************************/
#include <jendefs.h>
#include "nfc.h"

/****************************************************************************/
/***        Macro Definitions                                             ***/
/****************************************************************************/
/* OTA defines */
#define NFC_OTA_MAX_HEADER_SIZE               69 /*!< Maximum ZigBee format OTA header size */
#define NFC_OTA_MIN_HEADER_SIZE               56 /*!< Minimum ZigBee format OTA header size */
#define NFC_OTA_STRING_SIZE                   32 /*!< ZigBee OTA string size */
#define NFC_OTA_TAG_SIZE                       6 /*!< ZigBee OTA tag size */
#define NFC_OTA_FILE_IDENTIFIER       0x0BEEF11E /*!< ZigBee OTA file identifier */
#define NFC_OTA_TAG_ID_UPGRADE_IMAGE      0x0000 /*!< ZigBee OTA upgrade image tag ID */
#define NFC_OTA_BIN_HEADER_OFFSET          0x160 /*!< Offset of ZigBee header from start of binary */
#define NFC_OTA_BIN_CHECKSUM_OFFSET           69 /*!< Offset of JN checksum from start of ZigBee header */
#define NFC_OTA_BOOTLOADER_HEADER_SIZE       336 /*!< Bootloader header size */

/****************************************************************************/
/***        Type Definitions                                              ***/
/****************************************************************************/
/*! ZigBee OTA file header (69 bytes) */
typedef struct
{
    uint32 u32FileIdentifier;                   /*!< ( 4) ZigBee OTA file identifier */
    uint16 u16HeaderVersion;                    /*!< ( 2) ZigBee OTA file format version */
    uint16 u16HeaderLength;                     /*!< ( 2) ZigBee OTA header length */
    uint16 u16HeaderControlField;               /*!< ( 2) ZigBee OTA header flags */
    uint16 u16ManufacturerCode;                 /*!< ( 2) ZigBee image manufacturer code */
    uint16 u16ImageType;                        /*!< ( 2) ZigBee image ID */
    uint32 u32FileVersion;                      /*!< ( 4) ZigBee image version */
    uint16 u16StackVersion;                     /*!< ( 2) ZigBee stack version */
    uint8  stHeaderString[NFC_OTA_STRING_SIZE]; /*!< (32) ZigBee image string */
    uint32 u32TotalImage;                       /*!< ( 4) ZigBee total image size */
    uint8  u8SecurityCredVersion;               /*!< ( 1) ZigBee OTA security credentials */
    uint64 u64UpgradeFileDest;                  /*!< ( 8) ZigBee OTA file destination */
    uint16 u16MinHwVersion;                     /*!< ( 2) ZigBee OTA minimum hardware version */
    uint16 u16MaxHwVersion;                     /*!< ( 2) ZigBee OTA maximum hardware version */
} tsNfcOtaZbHeader;

/* Packed structures */
#pragma GCC diagnostic ignored "-Wpacked"
#pragma GCC diagnostic ignored "-Wattributes"
/*! NFC OTA Image Command (64 bytes) */
typedef struct
{
    uint8   u8Cmd;                               /*!< ( 1) OTA Command ID */
    uint8   u8Flags;                             /*!< ( 1) OTA Command Flags */
    uint32 u32FileIdentifier;                    /*!< ( 4) OTA File Identifier */
    uint16 u16ManufacturerCode;                  /*!< ( 2) OTA Manufacturer Code */
    uint16 u16ImageType;                         /*!< ( 2) OTA Image Type */
    uint16 u16StackVersion;                      /*!< ( 2) OTA Stack Version */
    uint8   stHeaderString[NFC_OTA_STRING_SIZE]; /*!< (32) OTA Header String (not terminated) */
    uint32 u32FileVersionActive;                 /*!< ( 4) OTA File Version Active */
    uint32 u32FileVersionInactive;               /*!< ( 4) OTA File Version Inactive */
    uint32 u32TotalImageInactive;                /*!< ( 4) OTA Image Size Inactive */
    uint32 u32SavedImageInactive;                /*!< ( 4) OTA Image Saved Inactive */
    uint32 u32OffsetImageInactive;               /*!< ( 4) OTA Offset Image Inactive */
} PACK tsNfcOtaCmdImage;

#define NFC_OTA_CMD_DATA_DATA_SIZE (NFC_SRAM_SIZE - 5) /*!< Data command data size */
/*! OTA Data Command (64 bytes) */
typedef struct
{
    uint8   u8Cmd;                              /*!< ( 1) OTA Command ID */
    uint32 u32Offset;                           /*!< ( 4) Offset of data within image */
    uint8  au8Data[NFC_OTA_CMD_DATA_DATA_SIZE]; /*!< (59) Image data */
} PACK tsNfcOtaCmdData;
#pragma GCC diagnostic pop

/****************************************************************************/
/***        Exported Functions                                            ***/
/****************************************************************************/
/****************************************************************************
 *
 *  NAME:  NFC_OTA_bZbHeaderFromBuffer
 */
/*! \brief Converts raw buffer data to ZigBee OTA header structure
 *
 ****************************************************************************/
PUBLIC bool_t NFC_OTA_bZbHeaderFromBuffer(
    tsNfcOtaZbHeader *psNfcOtaZbHeader,    /*!< ZigBee OTA header structure to populate (output) */
    uint32         *pu32ImageOffset,       /*!< OTA image offset for .ota header (NULL to skip) */
    uint32         *pu32Checksum,          /*!< OTA checksum for .bin header (NULL to skip) */
    uint8           *pu8Buffer,            /*!< Data buffer (input) */
    uint32           u32Length             /*!< Length of data in buffer (input) */
    );
/****************************************************************************
 *
 *  NAME:  NFC_OTA_vCmdImageToBuffer
 */
/*! \brief Converts tsNfcOtaCmdImage data structure to raw data buffer
 *
 ****************************************************************************/
PUBLIC void   NFC_OTA_vCmdImageToBuffer  (
    tsNfcOtaCmdImage  *psNfcOtaCmdImage,   /*!< Pointer to structure for input data */
    uint8            *pu8Buffer            /*!< Pointer to buffer for output data */
    );

/****************************************************************************
 *
 *  NAME:  NFC_OTA_vCmdImageFromBuffer
 */
/*! \brief Converts raw buffer data to tsNfcOtaCmdImage data structure
 *
 ****************************************************************************/
PUBLIC void   NFC_OTA_vCmdImageFromBuffer(
    tsNfcOtaCmdImage  *psNfcOtaCmdImage,   /*!< Pointer to structure for input data */
    uint8            *pu8Buffer            /*!< Pointer to buffer for output data */
    );

/****************************************************************************
 *
 *  NAME:  NFC_OTA_vCmdDataToBuffer
 */
/*! \brief Converts tsNfcOtaCmdData data structure to raw data buffer
 *
 ****************************************************************************/
PUBLIC void   NFC_OTA_vCmdDataToBuffer   (
    tsNfcOtaCmdData  *psNfcOtaCmdData,     /*!< Pointer to structure for input data */
    uint8           *pu8Buffer             /*!< Pointer to buffer for output data */
    );

/****************************************************************************
 *
 *  NAME:  NFC_OTA_vCmdDataFromBuffer
 */
/*! \brief Converts raw buffer data to tsNfcOtaCmdData data structure
 *
 ****************************************************************************/
PUBLIC void   NFC_OTA_vCmdDataFromBuffer (
    tsNfcOtaCmdData  *psNfcOtaCmdData,     /*!< Pointer to structure for input data */
    uint8           *pu8Buffer             /*!< Pointer to buffer for output data */
    );

/****************************************************************************
 *
 *  NAME:  NFC_OTA_bFlashReadBytes
 */
/*! \brief Reads bytes from flash
 *
 ****************************************************************************/
PUBLIC bool_t NFC_OTA_bFlashReadBytes(
    uint8 *pu8Buffer,   /*!< Pointer to buffer to read into */
    uint32 u32Address,  /*!< Flash byte address to read from */
    uint32 u32Length    /*!< Amount of data to read */
    );

/****************************************************************************
 *
 *  NAME:  NFC_OTA_bFlashWritePage
 */
/*! \brief Write page (512 bytes) to flash
 *
 ****************************************************************************/
PUBLIC bool_t NFC_OTA_bFlashWritePage(
    uint8 *pu8Buffer, /*!< Pointer to buffer to write from */
    uint32 u32Page    /*!< Flash page to write to */
    );

/****************************************************************************/
/***        Exported Variables                                            ***/
/****************************************************************************/

#endif /* NFC_NWK_H_ */
/****************************************************************************/
/***        END OF FILE                                                   ***/
/****************************************************************************/
