/*****************************************************************************
 * Copyright  2018 NXP
 * All rights reserved
 *

 * SPDX-License-Identifier: LA_OPT_BASE_LICENSE v24 May 2018
 *****************************************************************************/
/*****************************************************************************
 *
 * MODULE:      NFC
 *
 * COMPONENT:   ntag.h
 *
 * AUTHOR:      Martin Looker
 *
 * DESCRIPTION: NTAG driver for reading and writing data (interface)
 *
 * $HeadURL: https://www.collabnet.nxp.com/svn/lprf_sware/Projects/Components/NFC/Tags/+v1025/Include/ntag.h $
 *
 * $Revision: 102076 $
 *
 * $LastChangedBy: cnh01837 $
 *
 * $LastChangedDate: 2023-11-17 16:06:10 +0800 (周五, 17 11月 2023) $
 *
 * $Id: ntag.h 102076 2023-11-17 08:06:10Z cnh01837 $
 *
 ****************************************************************************
 *
 * This software is owned by NXP B.V. and/or its supplier and is protected
 * under applicable copyright laws. All rights are reserved. We grant You,
 * and any third parties, a license to use this software solely and
 * exclusively on NXP products [NXP Microcontrollers such as JN5168, JN5179].
 * You, and any third parties must reproduce the copyright and warranty notice
 * and any other legend of ownership on each copy or partial copy of the
 * software.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright NXP B.V. 2016-2018. All rights reserved
 *
 ***************************************************************************/
/*!
 * \file  ntag.h
 *
 * \brief NTAG driver for reading and writing data (interface)
 *
 * ntag.h contains low level APIs for reading and writing raw data to the NTAG.
 *
 * \warning Earlier versions of this module used an asynchronous model, the
 * read/write functions would request an action and some time later the final
 * result would be returned using a callback function. This implementation uses
 * a synchronous model where the read/write functions do not return until the
 * operation is complete.
 *
 * When writing more than one block (16 bytes) to EEPROM it is recommended to
 * split the write into blocks and allow 5ms between each so the CPU can
 * provide processing power to other modules (instead of using the busy waits
 * between each block that the write function enforces).
 *
 ***************************************************************************/
#ifndef NTAG_H_
#define NTAG_H_

/****************************************************************************/
/***        Include Files                                                 ***/
/****************************************************************************/
#include <jendefs.h>

/****************************************************************************/
/***        Macro Definitions                                             ***/
/****************************************************************************/
#define NTAG_BLOCK_SIZE                     16  /*!< NTAG block size */
#define NTAG_WRITE_BLOCK_EEPROM_DELAY_US  4500  /*!< Recommended delay after writing one block to EEPROM */
#define NTAG_WRITE_BLOCK_SRAM_DELAY_US     400  /*!< Recommended delay after writing one block to SRAM */

/****************************************************************************/
/***        Type Definitions                                              ***/
/****************************************************************************/

/****************************************************************************/
/***        Exported Functions (called by application)                    ***/
/****************************************************************************/
/****************************************************************************
 *
 *  NAME:  NTAG_vInitialise
 */
/*! \brief Initialises NTAG data and hardware
 *
 ****************************************************************************/
PUBLIC void NTAG_vInitialise(
       uint8    u8Address,     /*!< Reader I2C address (0xFF for automatic detection) */
       uint8    u8I2cScl,      /*!< 6x: 16   = DIO16
                                *       others = DIO14
                                *   7x: 4      = DIO4
                                *       others = DIO3
                                *   8x: 6      = I2C1 DIO6
                                *       12     = I2C1 DIO12
                                *       15     = I2C0 DIO15
                                *       0xff   = I2C2 (internal NTAG)
                                *       others = I2C0 DIO10 */
       uint32  u32FrequencyHz, /*!< Frequency in Hz */
       uint8    u8InputFd      /*!< Input DIO for field detect
                                *   0xff to ignore FD line */
       );

/****************************************************************************
 *
 *  NAME:  NTAG_vDeinitialise
 */
/*! \brief Deinitialises NTAG data and hardware
 *
 ****************************************************************************/
PUBLIC void NTAG_vDeinitialise(void);

/****************************************************************************
 *
 * NAME: NTAG_bReadFd
 */
/*! \brief Reads FD input
 *
 * \retval FALSE FD is low (in a field)
 * \retval TRUE  FD is high (not in a field)
 *
 ****************************************************************************/
 PUBLIC bool_t NTAG_bReadFd(void);

/****************************************************************************
 *
 * NAME: NTAG_bRead
 */
/*! \brief Requests blocking read of NTAG data
 *
 * Only operates on 16 byte blocks, so u32ReadAddress and u32ReadLength must
 * be multiples of 16.
 *
 * Does not return until the read is completed
 *
 * \retval TRUE  Read succeeded
 * \retval FALSE Read failed
 *
 ****************************************************************************/
PUBLIC  bool_t      NTAG_bRead(
        uint32      u32ReadAddress,     /*!< Byte address of read (multiple of 16) */
        uint32      u32ReadLength,      /*!< Number of bytes to read (multiple of 16) */
        uint8       *pu8ReadData        /*!< Buffer to read data into */
        );

/****************************************************************************
 *
 *  NAME:  NTAG_bReadVersion
 */
/*! \brief Requests blocking read of NTAG version data
 *
 * The version information is always 8 bytes in size the buffer and length
 * needs to be set appropriately. The header (first 16 bytes) must have been
 * read prior to calling this function and identified as an NXP NTAG.
 *
 * Does not return until the read is completed
 *
 * \retval TRUE  Read successful
 * \retval FALSE Read failed
 *
 ****************************************************************************/
PUBLIC  bool_t NTAG_bReadVersion(
        uint32    u32ReadLength, /*!< Number of bytes to read (minimum 8) */
        uint8    *pu8ReadData    /*!< Buffer to read data into */
        );

/****************************************************************************
 *
 *  NAME:  NTAG_bWrite
 */
/*! \brief Requests blocking write of NTAG data
 *
 * Only operates on 16 byte blocks, so u32WriteAddress and u32WriteLength must
 * be multiples of 16.
 *
 * Does not return until the write is completed
 *
 * Writing a block to SRAM takes 400us, writing a block to EEPROM takes 4500us.
 *
 * When writing more than one block appropriate delay values should be passed
 * in the u32WriteDelayUs parameter. Note that the delay is not applied before
 * the first block or after the last block of the data write, the calling layer
 * may need to implement its own delay if being used immediately before or
 * after another NTAG blocking write.
 *
 * \retval TRUE  Write suceeded
 * \retval FALSE Write failed
 *
 ****************************************************************************/
PUBLIC  bool_t NTAG_bWrite(
        uint32    u32WriteAddress,  /*!< Byte address of write */
        uint32    u32WriteLength,   /*!< Number of bytes to write */
        uint8    *pu8WriteData,     /*!< Buffer to write data from */
        uint32    u32WriteDelayUs   /*!< Delay between each block (us) */
        );

/****************************************************************************
 *
 *  NAME:  NTAG_bReadReg
 */
/*! \brief Requests blocking read of NTAG register data
 *
 * The register data is always 1 byte, the data length and buffer need to be
 * sized appropriately.
 *
 * Does not return until the read is completed
 *
 * \retval      TRUE  Read successful
 * \retval      FALSE Read failed
 *
 ****************************************************************************/
PUBLIC  bool_t NTAG_bReadReg(
        uint32    u32ReadAddress, /*<! Byte address of read */
        uint32    u32ReadLength,  /*<! Number of bytes to read (minimum 1) */
        uint8    *pu8ReadData     /*<! Buffer to read data into */
        );

/****************************************************************************
 *
 *  NAME:  NTAG_bWriteReg
 */
/*! \brief Requests blocking write of NTAG register data
 *
 * The register data is always 2 bytes, the data length and buffer need to be
 * sized appropriately. The first byte specifies a mask of bits to be written
 * to the register, the second byte specifies the values to set the bits
 * specified by the mask.
 *
 * Does not return until the write is completed
 *
 * \retval TRUE  Write successful
 * \retval FALSE Write failed
 *
 ****************************************************************************/
PUBLIC  bool_t NTAG_bWriteReg(
        uint32    u32WriteAddress, /*!< Byte address of write */
        uint32    u32WriteLength,  /*!< Number of bytes to write */
        uint8    *pu8WriteData     /*!< Buffer to write data from: [0] = RegMask, [1] = RegValue */
        );

/****************************************************************************
 *
 *  NAME:  NTAG_u32Ntag
 */
/*! \brief Returns the NTAG model
 *
 * The NTAG_bReadVersion() function must have been called previously in
 * order to detect the NTAG model.
 *
 * \retval NFC_NTAG_UNKNOWN  Unknown NTAG model
 * \retval NFC_NTAG_NT3H1101 NT3H1101 NTAG model
 * \retval NFC_NTAG_NT3H1201 NT3H1201 NTAG model
 * \retval NFC_NTAG_NT3H2111 NT3H2111 NTAG model
 * \retval NFC_NTAG_NT3H2211 NT3H2211 NTAG model
 *
 ****************************************************************************/
PUBLIC  uint32 NTAG_u32Ntag(void);

/****************************************************************************
 *
 *  NAME:  NTAG_u32Config
 */
/*! \brief Returns the byte address of the NTAG's configuration registers
 *
 * The NTAG_bReadVersion() function must have been called previously in
 * order to determine the address.
 *
 * \returns Byte address of configuration registers
 *
 ****************************************************************************/
PUBLIC  uint32 NTAG_u32Config(void);

/****************************************************************************
 *
 *  NAME:        NTAG_u32Session
 */
/*! \brief Returns the byte address of the NTAG's session registers
 *
 * The NTAG_bReadVersion() function must have been called previously in
 * order to determine the address.
 *
 * \returns Byte address of session registers
 *
 ****************************************************************************/
PUBLIC  uint32 NTAG_u32Session(void);

/****************************************************************************
 *
 *  NAME:        NTAG_u32Sram
 */
/*! \brief Returns the byte address of the NTAG's SRAM
 *
 * The NTAG_bReadVersion() function must have been called previously in
 * order to determine the address.
 *
 * \returns Byte address of SRAM
 *
 ****************************************************************************/
PUBLIC  uint32 NTAG_u32Sram(void);

/****************************************************************************
 *
 *  NAME:        NTAG_pu8Header
 */
/*! \brief Returns the 16 header bytes of the NTAG's memory
 *
 * The NTAG_bRead() function must have been called previously with a read of
 * these first 16 bytes from address 0.
 *
 * \returns Pointer to header data
 *
 ****************************************************************************/
PUBLIC  uint8 *NTAG_pu8Header(void);

/****************************************************************************
 *
 *  NAME:        NTAG_pu8Version
 */
/*! \brief Returns the 8 version information bytes for the NTAG
 *
 * The NTAG_bReadVersion() function must have been called previously.
 *
 * \returns Pointer to version data
 *
 ****************************************************************************/
PUBLIC  uint8 *NTAG_pu8Version(void);

/****************************************************************************/
/***        Exported Variables                                            ***/
/****************************************************************************/

#endif /* NTAG_H_ */
/****************************************************************************/
/***        END OF FILE                                                   ***/
/****************************************************************************/
